# coding: utf-8
"""Numeric tools."""
from __future__ import annotations

import numpy as np

from monty.collections import dict2namedtuple
from abipy.tools import duck


def build_mesh(x0: float, num: int, step: float, direction: str) -> tuple[list, int]:
    """
    Generate a linear mesh of step `step` that is centered on x0 if
    directions == "centered" or a mesh that starts/ends at x0 if direction is `>`/`<`.
    Return mesh and index of x0.
    """
    if direction == "centered":
        start = x0 - num * step
        return [start + i * step for i in range(2 * num + 1)], num

    elif direction in (">", "<"):
        start, ix0 = x0, 0
        if direction == "<":
            ix0 = num - 1
            step = -abs(step)

        return sorted([start + i * step for i in range(num)]), ix0

    raise ValueError(f"Invalid {direction=}")


def transpose_last3dims(arr) -> np.ndarray:
    """
    Transpose the last three dimensions of arr: (...,x,y,z) --> (...,z,y,x).
    """
    axes = np.arange(arr.ndim)
    axes[-3:] = axes[::-1][:3]

    view = np.transpose(arr, axes=axes)
    return np.ascontiguousarray(view)


def add_periodic_replicas(arr: np.ndarray) -> np.ndarray:
    """
    Returns a new array of shape=(..., nx+1,ny+1,nz+1) with redundant data points.

    Periodicity in enforced only on the last three dimensions.
    """
    ishape, ndim = np.array(arr.shape), arr.ndim
    oshape = ishape.copy()

    if ndim == 1:
        oarr = np.empty(ishape+1, dtype=arr.dtype)
        oarr[:-1] = arr
        oarr[-1] = arr[0]

    elif ndim == 2:
        oarr = np.empty(oshape+1, dtype=arr.dtype)
        oarr[:-1,:-1] = arr
        oarr[-1,:-1] = arr[0,:]
        oarr[:-1,-1] = arr[:,0]
        oarr[-1,-1] = arr[0,0]

    else:
        # Add periodic replica along the last three directions.
        oshape[-3:] = oshape[-3:] + 1
        oarr = np.empty(oshape, dtype=arr.dtype)

        for x in range(ishape[-3]):
            for y in range(ishape[-2]):
                oarr[..., x, y, :-1] = arr[..., x, y, :]
                oarr[..., x, y, -1] = arr[..., x, y, 0]

            oarr[..., x, y + 1, :-1] = arr[..., x, 0, :]
            oarr[..., x, y + 1, -1] = arr[..., x, 0, 0]

        oarr[..., x + 1, :, :] = oarr[..., 0, :, :]

    return oarr


def data_from_cplx_mode(cplx_mode: str, arr, tol=None):
    """
    Extract the data from the numpy array ``arr`` depending on the values of ``cplx_mode``.

    Args:
        cplx_mode: Possible values in ("re", "im", "abs", "angle")
            "re" for the real part,
            "im" for the imaginary part.
            "all" for both re and im.
            "abs" means that the absolute value of the complex number is shown.
            "angle" will display the phase of the complex number in radians.
        tol: If not None, values below tol are set to zero. Cannot be used with "angle"
    """
    if cplx_mode == "re":
        val = arr.real
    elif cplx_mode == "im":
        val = arr.imag
    elif cplx_mode == "all":
        val = arr
    elif cplx_mode == "abs":
        val = np.abs(arr)
    elif cplx_mode == "angle":
        val = np.angle(arr, deg=False)
        if tol is not None:
            raise ValueError("Tol cannot be used with cplx_mode = angle")
    else:
        raise ValueError("Unsupported mode `%s`" % str(cplx_mode))

    return val if tol is None else np.where(np.abs(val) > tol, val, 0)


def is_diagonal(matrix, atol=1e-12):
    """
    Return True if matrix is diagonal.
    """
    m = matrix.copy()
    np.fill_diagonal(m, 0)

    if issubclass(matrix.dtype.type, np.integer):
        return np.all(m == 0)
    else:
        return np.all(np.abs(m) <= atol)


#########################################################################################
# Tools to facilitate iterations
#########################################################################################


def alternate(*iterables):
    """
    [a[0], b[0], ... , a[1], b[1], ..., a[n], b[n] ...]
    >>> alternate([1,4], [2,5], [3,6])
    [1, 2, 3, 4, 5, 6]
    """
    items = []
    for tup in zip(*iterables):
        items.extend([item for item in tup])
    return items


def iflat(iterables):
    """
    Iterator over all elements of a nested iterable. It's recursive!

    >>> list(iflat([[0], [1,2, [3,4]]]))
    [0, 1, 2, 3, 4]
    """
    for item in iterables:
        if not hasattr(item, "__iter__"):
            yield item
        else:
            # iterable object.
            for it in iflat(item):
                yield it


def grouper(n, iterable, fillvalue=None):
    """
    >>> assert grouper(3, "ABCDEFG", "x") == [('A', 'B', 'C'), ('D', 'E', 'F'), ('G', 'x', 'x')]
    >>> assert grouper(3, [1, 2, 3, 4]) == [(1, 2, 3), (4, None, None)]
    """
    # https://stackoverflow.com/questions/434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks/434411#434411
    try:
        from itertools import zip_longest
    except ImportError:
        from itertools import izip_longest as zip_longest

    args = [iter(iterable)] * n
    return list(zip_longest(fillvalue=fillvalue, *args))


def sort_and_groupby(items, key=None, reverse=False, ret_lists=False):
    """
    Sort ``items`` using ``key`` function and invoke itertools.groupby to group items.
    If ret_lists is True, a tuple of lists (keys, groups) is returned else iterator.
    See itertools.groupby for further info.

    >>> sort_and_groupby([1, 2, 1], ret_lists=True)
    ([1, 2], [[1, 1], [2]])
    """
    from itertools import groupby
    if not ret_lists:
        return groupby(sorted(items, key=key, reverse=reverse), key=key)
    else:
        keys, groups = [], []
        for hvalue, grp in groupby(sorted(items, key=key, reverse=reverse), key=key):
            keys.append(hvalue)
            groups.append(list(grp))
        return keys, groups


#########################################################################################
# Sorting and ordering
#########################################################################################

def prune_ord(alist: list) -> list:
    """
    Return new list where all duplicated items in alist are removed

    1) The order of items in alist is preserved.
    2) items in alist MUST be hashable.

    Taken from http://code.activestate.com/recipes/52560/
    >>> prune_ord([1, 1, 2, 3, 3])
    [1, 2, 3]
    """
    mset = {}
    return [mset.setdefault(e, e) for e in alist if e not in mset]

#########################################################################################
# Special functions
#########################################################################################


def gaussian(x, width, center=0.0, height=None):
    """
    Returns the values of gaussian(x) where x is array-like.

    Args:
        x: Input array.
        width: Width of the gaussian.
        center: Center of the gaussian.
        height: height of the gaussian. If height is None, a normalized gaussian is returned.
    """
    x = np.asarray(x)
    if height is None: height = 1.0 / (width * np.sqrt(2 * np.pi))

    return height * np.exp(-((x - center) / width) ** 2 / 2.)


def lorentzian(x, width, center=0.0, height=None):
    """
    Returns the values of gaussian(x) where x is array-like.

    Args:
        x: Input array.
        width: Width of the Lorentzian (half-width at half-maximum)
        center: Center of the Lorentzian.
        height: height of the Lorentzian. If height is None, a normalized Lorentzian is returned.
    """
    x = np.asarray(x)
    if height is None: height = 1.0 / (width * np.pi)

    return height * width**2 / ((x - center) ** 2 + width ** 2)

#=====================================
# === Data Interpolation/Smoothing ===
#=====================================


def smooth(x, window_len=11, window='hanning'):
    """
    smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    Taken from http://www.scipy.org/Cookbook/SignalSmooth

    Args:
        x:
            the input signal
        window_len:
            the dimension of the smoothing window. it should be an odd integer
        window:
            the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'.
            'flat' window will produce a moving average smoothing.

    Returns:
        the smoothed signal.

    example::

        t = linspace(-2,2,0.1)
        x = sin(t)+randn(len(t))*0.1
        y = smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window_len % 2 == 0:
        raise ValueError("window_len should be odd.")

    windows = ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']

    if window not in windows:
        raise ValueError("window must be in: " + str(windows))

    s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]

    if window == 'flat': # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')

    s = window_len // 2
    e = s + len(x)
    return y[s:e]


def find_convindex(values, tol, min_numpts=1, mode="abs", vinf=None):
    """
    Given a list of values and a tolerance tol, returns the leftmost index for which

        abs(value[i] - vinf) < tol if mode == "abs"

    or
        abs(value[i] - vinf) / vinf < tol if mode == "rel"

    Args:
        tol: Tolerance
        min_numpts: Minimum number of points that must be converged.
        mode: "abs" for absolute convergence, "rel" for relative convergence.
        vinf: Used to specify an alternative value instead of values[-1].
            By default, vinf = values[-1]

    Return:
        -1 if convergence is not achieved else the index in values.
    """
    vinf = values[-1] if vinf is None else vinf

    if mode == "abs":
        vdiff = [abs(v - vinf) for v in values]
    elif mode == "rel":
        vdiff = [abs(v - vinf) / vinf for v in values]
    else:
        raise ValueError("Wrong mode %s" % mode)

    numpts, i = len(vdiff), -2
    if numpts > min_numpts and vdiff[-2] < tol:
        for i in range(numpts-1, -1, -1):
            if vdiff[i] > tol:
                break
        if (numpts - i - 1) < min_numpts: i = -2

    return i + 1


class BlochRegularGridInterpolator:
    """
    This object interpolates the periodic part of a Bloch state in real space.
    """

    def __init__(self, structure, datar, add_replicas=True):
        """
        Args:
            structure: :class:`Structure` object.
            datar: [ndt, nx, ny, nz] array.
            add_replicas: If True, data is padded with redundant data points.
                in order to have a periodic 3D array of shape=[ndt, nx+1, ny+1, nz+1].
        """
        from scipy.interpolate import RegularGridInterpolator
        self.structure = structure

        if add_replicas:
            datar = add_periodic_replicas(datar)

        self.dtype = datar.dtype
        # We want a 4d array (ndt arrays of shape (nx, ny, nz)
        nx, ny, nz = datar.shape[-3:]
        datar = np.reshape(datar, (-1,) + (nx, ny, nz))
        self.ndt = len(datar)
        x = np.linspace(0, 1, num=nx)
        y = np.linspace(0, 1, num=ny)
        z = np.linspace(0, 1, num=nz)

        # Build `ndt` interpolators. Note that RegularGridInterpolator supports
        # [nx, ny, nz, ...] arrays but then each call operates on the full set of
        # ndt components and this complicates the declation of callbacks
        # operating on a single component.
        self._interpolators = [None] * self.ndt
        for i in range(self.ndt):
            self._interpolators[i] = RegularGridInterpolator((x, y, z), datar[i])

    def eval_line(self, point1, point2, num=200, cartesian=False, kpoint=None):
        """
        Interpolate values along a line.

        Args:
            point1: First point of the line. Accepts 3d vector or integer.
                The vector is in reduced coordinates unless `cartesian == True`.
                If integer, the first point of the line is given by the i-th site of the structure
                e.g. `point1=0, point2=1` gives the line passing through the first two atoms.
            point2: Second point of the line. Same API as `point1`.
            num: Number of points sampled along the line.
            cartesian: By default, `point1` and `point1` are interpreted as points in fractional
                coordinates (if not integers). Use True to pass points in cartesian coordinates.
            kpoint: k-point in reduced coordinates. If not None, the phase-factor e^{ikr} is included.

        Return: named tuple with
            site1, site2: None if the points do not represent atomic sites.
            points: Points in fractional coords.
            dist: the distance of points along the line in Ang.
            values: numpy array of shape [ndt, num] with interpolated values.
        """
        site1 = None
        if duck.is_intlike(point1):
            if point1 > len(self.structure):
                raise ValueError("point1: %s > natom: %s" % (point1, len(self.structure)))
            site1 = self.structure[point1]
            point1 = site1.coords if cartesian else site1.frac_coords

        site2 = None
        if duck.is_intlike(point2):
            if point2 > len(self.structure):
                raise ValueError("point2: %s > natom: %s" % (point2, len(self.structure)))
            site2 = self.structure[point2]
            point2 = site2.coords if cartesian else site2.frac_coords

        point1 = np.reshape(point1, (3,))
        point2 = np.reshape(point2, (3,))
        if cartesian:
            red_from_cart = self.structure.lattice.inv_matrix.T
            point1 = np.dot(red_from_cart, point1)
            point2 = np.dot(red_from_cart, point2)

        p21 = point2 - point1
        line_points = np.reshape([alpha * p21 for alpha in np.linspace(0, 1, num=num)], (-1, 3))
        dist = self.structure.lattice.norm(line_points)
        line_points += point1

        return dict2namedtuple(site1=site1, site2=site2, points=line_points, dist=dist,
                               values=self.eval_points(line_points, kpoint=kpoint))

    def eval_points(self, frac_coords, idt=None, cartesian=False, kpoint=None):
        """
        Interpolate values on an arbitrary list of points.

        Args:
            frac_coords: List of points in reduced coordinates unless `cartesian`.
            idt: Index of the sub-array to interpolate. If None, all sub-arrays are interpolated.
            cartesian: True if points are in cartesian coordinates.
            kpoint: k-point in reduced coordinates. If not None, the phase-factor e^{ikr} is included.

        Return:
            [ndt, npoints] array or [1, npoints] if idt is not None
        """
        frac_coords = np.reshape(frac_coords, (-1, 3))
        if cartesian:
            red_from_cart = self.structure.lattice.inv_matrix.T
            frac_coords = [np.dot(red_from_cart, v) for v in frac_coords]

        uc_coords = np.reshape(frac_coords, (-1, 3)) % 1

        if idt is None:
            values = np.empty((self.ndt, len(uc_coords)), dtype=self.dtype)
            for idt in range(self.ndt):
                values[idt] = self._interpolators[idt](uc_coords)
        else:
            values = self._interpolators[idt](uc_coords)

        if kpoint is not None:
            if hasattr(kpoint, "frac_coords"): kpoint = kpoint.frac_coords
            kpoint = np.reshape(kpoint, (3,))
            values *= np.exp(2j * np.pi * np.dot(frac_coords, kpoint))

        return values


def find_degs_sk(enesb, atol):
    """
    Return list of lists with the indices of the degenerated bands.

    Args:
        enesb: Iterable with energies for the different bands.
            Energies are assumed to be ordered.
        atol: Absolute tolerance. Two states are degenerated if they differ by less than `atol`.

    Return:
        List of lists. The i-th item contains the indices of the degenerates states
            for the i-th degenerated set.

    :Examples:

    >>> find_degs_sk([1, 1, 2, 3.4, 3.401], atol=0.01)
    [[0, 1], [2], [3, 4]]
    """
    ndeg = 0
    degs = [[0]]
    e0 = enesb[0]
    for ib, ee in enumerate(enesb[1:]):
        ib += 1
        if abs(ee - e0) > atol:
            e0 = ee
            ndeg += 1
            degs.append([ib])
        else:
            degs[ndeg].append(ib)

    return degs



#!/usr/bin/env python
r"""
Starting a Flow from external files
===================================

This example shows how to build a flow for NSCF calculations
in which the dependency is given by an external density file.
"""
import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_scf_nscf_inputs(paral_kgb=1):
    """Returns two input files: GS run and NSCF on a high symmetry k-mesh."""
    multi = abilab.MultiDataset(structure=abidata.cif_file("si.cif"),
                                pseudos=abidata.pseudos("14si.pspnc"), ndtset=2)

    # Global variables
    ecut = 6
    global_vars = dict(
        ecut=ecut,
        nband=8,
        nstep=15,
        paral_kgb=paral_kgb,
    )

    if multi.ispaw:
        global_vars.update(pawecutdg=2*ecut)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[8, 8, 8], shiftk=[0, 0, 0])
    multi[0].set_vars(tolvrs=1e-6)

    # Dataset 2 (NSCF run)
    kptbounds = [
        [0.5, 0.0, 0.0], # L point
        [0.0, 0.0, 0.0], # Gamma point
        [0.0, 0.5, 0.5], # X point
    ]

    multi[1].set_kpath(ndivsm=6, kptbounds=kptbounds)
    multi[1].set_vars(tolwfr=1e-12)

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input = multi.split_datasets()

    return scf_input, nscf_input


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Get the SCF and the NSCF input.
    scf_input, nscf_input = make_scf_nscf_inputs()

    # Build the flow.
    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Create a Work, all tasks in work will start from the DEN file.
    # Note that the file must exist when the work is created
    # Use the standard approach based on tasks and works if
    # there's a node who needs a file produced in the future.
    work = flow.new_work()
    den_filepath = abidata.ref_file("si_DEN.nc")
    work.register_nscf_task(nscf_input, deps={den_filepath: "DEN"})

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Objects used to deal with symmetry operations in crystals."""
from __future__ import annotations

import sys
import abc
import warnings
import collections
import numpy as np
import pandas as pd
import spglib

from monty.string import is_string
from monty.itertools import iuptri
from monty.functools import lazy_property
from monty.termcolor import cprint
from monty.collections import dict2namedtuple
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from abipy.core.kpoints import wrap_to_ws, issamek, has_timrev_from_kptopt
from abipy.core.mixins import SlotPickleMixin


__all__ = [
    "LatticeRotation",
    "AbinitSpaceGroup",
]


def wrap_in_ucell(x):
    """
    Transforms x in its corresponding reduced number in the interval [0,1[."
    """
    return x % 1


def is_integer(x, atol=1e-08) -> bool:
    """
    True if all x is integer within the absolute tolerance atol.

    >>> is_integer([1., 2.])
    True
    >>> is_integer(1.01, atol=0.011)
    True
    >>> is_integer([1.01, 2])
    False
    """
    int_x = np.around(x)
    return np.allclose(int_x, x, atol=atol)


def mati3inv(mat3, trans=True):
    """
    Invert and transpose orthogonal 3x3 matrix of INTEGER elements.

    Args:
        mat3: (3, 3) matrix-like object with integer elements

    Returns:
        |numpy-array| with the TRANSPOSE of the inverse of mat3 if trans==True.
        If trans==False, the inverse of mat3 is returned.

    .. note::

       Used for symmetry operations. This function applies to *ORTHOGONAL* matrices only.
       Since these form a group, inverses are also integer arrays.
    """
    mat3 = np.reshape(np.array(mat3, dtype=int), (3, 3))

    mit = np.empty((3, 3), dtype=int)
    mit[0,0] = mat3[1,1] * mat3[2,2] - mat3[2,1] * mat3[1,2]
    mit[1,0] = mat3[2,1] * mat3[0,2] - mat3[0,1] * mat3[2,2]
    mit[2,0] = mat3[0,1] * mat3[1,2] - mat3[1,1] * mat3[0,2]
    mit[0,1] = mat3[2,0] * mat3[1,2] - mat3[1,0] * mat3[2,2]
    mit[1,1] = mat3[0,0] * mat3[2,2] - mat3[2,0] * mat3[0,2]
    mit[2,1] = mat3[1,0] * mat3[0,2] - mat3[0,0] * mat3[1,2]
    mit[0,2] = mat3[1,0] * mat3[2,1] - mat3[2,0] * mat3[1,1]
    mit[1,2] = mat3[2,0] * mat3[0,1] - mat3[0,0] * mat3[2,1]
    mit[2,2] = mat3[0,0] * mat3[1,1] - mat3[1,0] * mat3[0,1]

    dd = mat3[0,0] * mit[0,0] + mat3[1,0] * mit[1,0] + mat3[2,0] * mit[2,0]

    # Make sure matrix is not singular
    if dd == 0:
        raise ValueError("Attempting to invert integer array: %s\n ==> determinant is zero." % str(mat3))

    mit = mit // dd
    if trans:
        return mit
    else:
        return mit.T.copy()


def _get_det(mat) -> float:
    """
    Return the determinant of a 3x3 rotation matrix mat.

    raises:
        ValueError if abs(det) != 1.
    """
    det = mat[0,0] * (mat[1,1] * mat[2,2] - mat[1,2] * mat[2,1])\
        - mat[0,1] * (mat[1,0] * mat[2,2] - mat[1,2] * mat[2,0])\
        + mat[0,2] * (mat[1,0] * mat[2,1] - mat[1,1] * mat[2,0])

    if abs(det) != 1:
        raise ValueError("Determinant must be +-1 while it is %s" % det)

    return det


def indsym_from_symrel(symrel, tnons, structure, tolsym=1e-8):
    r"""
    For each symmetry operation, find the number of the position to
    which each atom is sent in the unit cell by the INVERSE of the
    symmetry operation inv(symrel); i.e. this is the atom which, when acted
    upon by the given symmetry element isym, gets transformed into atom iatom.
    indirect indexing array for atoms, see symatm.F90.

    $ R^{-1} (xred(:,iat) - \tau) = xred(:,iat_sym) + R_0 $
    * indsym(4,  isym,iat) gives iat_sym in the original unit cell.
    * indsym(1:3,isym,iat) gives the lattice vector $R_0$.

    Args:
        symrel: int (nsym,3,3) array with real space symmetries expressed in reduced coordinates.
        tnons: float (nsym, 3) array with nonsymmorphic translations for each symmetry.
        structure: |Structure| object.
        tolsym: tolerance for the symmetries
    """
    natom = len(structure)
    nsym = len(symrel)
    xred = np.array([site.frac_coords for site in structure], dtype=float)
    typat = {i: site.specie.symbol for i, site in enumerate(structure)}

    rm1_list = np.empty_like(symrel)
    for isym in range(nsym):
        rm1_list[isym] = mati3inv(symrel[isym], trans=False)

    # Start testmn out at large value
    testmn = 1000000
    err = 0.0
    indsym = np.empty((natom, nsym, 4))

    # Implementation is similar to Abinit routine (including the order of the loops)
    for isym in range(nsym):
        for iatom in range(natom):
            tratm = np.matmul(rm1_list[isym], xred[iatom] - tnons[isym])
            # Loop through atoms, when types agree, check for agreement after primitive translation
            for jatm in range(natom):
                if typat[jatm] != typat[iatom]: continue
                test_vec = tratm - xred[jatm]
                # Find nearest integer part of difference
                trans = np.rint(test_vec)
                # Check whether, after translation, they agree
                test_vec = test_vec - trans
                diff = np.abs(test_vec).sum()
                # Abinit uses 1e-10 but python seems to require a slightly larger value.
                #if diff < 1e-10:
                if diff < 1e-9:
                    difmin = test_vec
                    indsym[iatom, isym, :3] = trans
                    indsym[iatom, isym, 3] = jatm
                    # Break out of loop when agreement is within tolerance
                    break
                else:
                    # Keep track of smallest difference if greater than tol10
                    if diff < testmn:
                        testmn = diff
                        # Note that abs() is not taken here
                        difmin = test_vec
                        indsym[iatom, isym, :3] = trans
                        indsym[iatom, isym, 3] = jatm

        # Keep track of maximum difference between transformed coordinates and nearest "target" coordinate
        difmax = np.abs(difmin).max()
        err = max(err, difmax)
        if difmax > tolsym:
            cprint(f"""
Trouble finding symmetrically equivalent atoms.
Applying inverse of symm number {isym} to atom number {iatom} of typat',typat(iatom) gives tratom=',tratom(1:3)
This is further away from every atom in crystal than the allowed tolerance.
The inverse symmetry matrix is',symrec(1,1:3,isym),ch10,&
                               ',symrec(2,1:3,isym),ch10,&
                               ',symrec(3,1:3,isym)
and the nonsymmorphic transl. tnons =',(tnons(mu,isym),mu=1,3)
The nearest coordinate differs by',difmin(1:3) for indsym(nearest atom)=',indsym(4,isym,iatom)

This indicates that when symatm attempts to find atoms symmetrically
related to a given atom, the nearest candidate is further away than some tolerance.
Should check atomic coordinates and symmetry group input data.
""", color="red")

    if err > tolsym:
        raise ValueError("maximum err %s is larger than tolsym: %s" % (err, tolsym))

    return indsym


class Operation(metaclass=abc.ABCMeta):
    """
    Abstract base class that defines the methods that must be
    implemented by the concrete class representing some sort of operation
    """
    @abc.abstractmethod
    def __eq__(self, other):
        """O1 == O2"""

    def __ne__(self, other):
        return not (self == other)

    @abc.abstractmethod
    def __mul__(self, other):
        """O1 * O2"""

    @abc.abstractmethod
    def __hash__(self):
        """Operation can be used as dictionary keys."""

    @abc.abstractmethod
    def inverse(self):
        """Returns the inverse of self."""

    def opconj(self, other):
        """Returns X^-1 S X where X is the other symmetry operation."""
        return other.inverse() * self * other

    @abc.abstractproperty
    def isE(self):
        """True if self is the identity operator"""

    #def commute(self, other)
    #    return self * other == other * self

    #def commutator(self, other)
    #    return self * other - other * self

    #def anticommute(self, other)
    #    return self * other == - other * self

    #def direct_product(self, other)


class SymmOp(Operation, SlotPickleMixin):
    """
    Crystalline symmetry.
    """
    _ATOL_TAU = 1e-8

    __slots__ = [
        "rot_r",
        "rotm1_r",
        "tau",
        "time_sign",
        "afm_sign",
        "rot_g",
        "_det",
        "_trace",
    ]

    # TODO: Add lattice?
    def __init__(self, rot_r, tau, time_sign, afm_sign, rot_g=None):
        """
        This object represents a space group symmetry i.e. a symmetry of the crystal.

        Args:
            rot_r: (3,3) integer matrix with the rotational part in real space in reduced coordinates (C order).
            tau: fractional translation in reduced coordinates.
            time_sign: -1 if time reversal can be used, +1 otherwise.
            afm_sign: anti-ferromagnetic part [+1, -1].
        """
        rot_r = np.asarray(rot_r)

        # Store R and R^{-1} in real space.
        self.rot_r, self.rotm1_r = rot_r, mati3inv(rot_r, trans=False)
        self.tau = np.asarray(tau)

        self.afm_sign, self.time_sign = afm_sign, time_sign
        assert afm_sign in [-1, 1] and time_sign in [-1, 1]

        # Compute symmetry matrix in reciprocal space: S = R^{-1t}
        if rot_g is None:
            self.rot_g = mati3inv(rot_r, trans=True)
        else:
            assert np.all(rot_g == mati3inv(rot_r, trans=True))
            self.rot_g = rot_g

    # operator protocol.
    def __eq__(self, other):
        # Note the two fractional traslations are equivalent if they differ by a lattice vector.
        return (np.all(self.rot_r == other.rot_r) and
                is_integer(self.tau - other.tau, atol=self._ATOL_TAU) and
                self.afm_sign == other.afm_sign and
                self.time_sign == other.time_sign)

    def __mul__(self, other):
        """
        Returns a new :class:`SymmOp` which is equivalent to apply the "other" :class:`SymmOp`
        followed by this one i.e:

        {R,t} {S,u} = {RS, Ru + t}
        """
        return self.__class__(rot_r=np.dot(self.rot_r, other.rot_r),
                              tau=self.tau + np.dot(self.rot_r, other.tau),
                              time_sign=self.time_sign * other.time_sign,
                              afm_sign=self.afm_sign * other.afm_sign)

    def __hash__(self):
        """
        :class:`Symmop` can be used as keys in dictionaries.
        Note that the hash is computed from integer values.
        """
        return int(8 * self.trace + 4 * self.det + 2 * self.time_sign)

    def inverse(self):
        """Returns inverse of transformation i.e. {R^{-1}, -R^{-1} tau}."""
        return self.__class__(rot_r=self.rotm1_r,
                              tau=-np.dot(self.rotm1_r, self.tau),
                              time_sign=self.time_sign,
                              afm_sign=self.afm_sign)

    @lazy_property
    def isE(self) -> bool:
        """True if identity operator."""
        return (np.all(self.rot_r == np.eye(3, dtype=int)) and
                is_integer(self.tau, atol=self._ATOL_TAU) and
                self.time_sign == 1 and
                self.afm_sign == 1)
    # end operator protocol.

    #@lazy_property
    #def order(self):
    #    """Order of the operation."""
    #    n = 0
    #    o = self
    #    while m < 1000:
    #        if o.isE: return n
    #        n += 1
    #        o = self * o
    #    else:
    #        raise ValueError("Cannot find order")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        def vec2str(vec):
            return "%2d,%2d,%2d" % tuple(v for v in vec)

        s = ""
        for i in range(3):
            s += "[" + vec2str(self.rot_r[i]) + ", %.3f]  " % self.tau[i] + "[" + vec2str(self.rot_g[i]) + "] "
            if i == 2:
                s += ", time_sign = %+1d, afm_sign = %+1d, det = %+1d" % (self.time_sign, self.afm_sign, self.det)
            s += "\n"

        return s

    @lazy_property
    def is_symmorphic(self) -> bool:
        """True if the fractional translation is non-zero."""
        return np.any(np.abs(self.tau) > 0.0)

    @lazy_property
    def det(self) -> int:
        """Determinant of the rotation matrix [-1, +1]."""
        return _get_det(self.rot_r)

    @lazy_property
    def trace(self) -> int:
        """Trace of the rotation matrix."""
        return self.rot_r.trace()

    @lazy_property
    def is_proper(self) -> bool:
        """True if the rotational part has determinant == 1."""
        return self.det == +1

    @lazy_property
    def has_timerev(self) -> bool:
        """True if symmetry contains the time-reversal operator."""
        return self.time_sign == -1

    @lazy_property
    def is_fm(self) -> bool:
        """True if self if ferromagnetic symmetry."""
        return self.afm_sign == +1

    @lazy_property
    def is_afm(self) -> bool:
        """True if self if anti-ferromagnetic symmetry."""
        return self.afm_sign == -1

    def rotate_k(self, frac_coords, wrap_tows=False):
        """
        Apply the symmetry operation to the k-point given in reduced coordinates.

        Sk is wrapped to the first Brillouin zone if wrap_tows is True.
        """
        sk = np.dot(self.rot_g, frac_coords) * self.time_sign

        return wrap_to_ws(sk) if wrap_tows else sk

    def preserve_k(self, frac_coords, ret_g0=True):
        """
        Check if the operation preserves the k-point modulo a reciprocal lattice vector.

        Args:
            frac_coords: Fractional coordinates of the k-point
            ret_g0: False if only the boolean result is wanted.

        Returns:
            bool, g0 = S(k) - k

            bool is True is self preserves k and g0 is an integer vector.
        """
        sk = self.rotate_k(frac_coords, wrap_tows=False)

        if ret_g0:
            return issamek(sk, frac_coords), np.array(np.round(sk - frac_coords), dtype=int)
        else:
            return issamek(sk, frac_coords)

    def rotate_r(self, frac_coords, in_ucell=False):
        """
        Apply the symmetry operation to a point in real space given in reduced coordinates.

        .. NOTE::

            We use the convention: symmop(r) = R^{-1] (r - tau)
        """
        rotm1_rmt = np.dot(self.rotm1_r, frac_coords - self.tau)

        return wrap_in_ucell(rotm1_rmt) if in_ucell else rotm1_rmt


class OpSequence(collections.abc.Sequence):
    """
    Mixin class providing the basic method that are common to containers of operations.
    """
    def __len__(self):
        return len(self._ops)

    def __iter__(self):
        return self._ops.__iter__()

    def __getitem__(self, slice):
        return self._ops[slice]

    def __contains__(self, op):
        return op in self._ops

    def __eq__(self, other):
        """
        Equality test.

        .. warning::

            The order of the operations in self and  in other is not relevant.
        """
        if other is None: return False
        if len(self) != len(other):
            return False

        # Check if each operation in self is also present in other.
        # The order is irrelevant.
        founds = []
        for i, op in enumerate(self):
            if op not in other: return False
            founds.append(i)

        if len(set(founds)) == len(founds):
            return True

        warnings.warn("self contains duplicated ops! Likely a bug!")
        return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        lines = [str(op) for op in self]
        return "\n".join(lines)

    def show_ops(self, stream=sys.stdout) -> None:
        lines = [str(op) for op in self]
        stream.writelines("\n".join(lines))

    def count(self, op) -> int:
        """Returns the number of occurences of operation op in self."""
        return self._ops.count(op)

    def index(self, op) -> int:
        """
        Return the (first) index of operation op in self.

        Raises:
            ValueError if not found.
        """
        return self._ops.index(op)

    def find(self, op) -> int:
        """Return the (first) index of op in self. -1 if not found."""
        try:
            return self.index(op)
        except ValueError:
            return -1

    def is_group(self) -> bool:
        """True if this set of operations represent a group."""
        check = 0

        # Identity must be present.
        if [op.isE for op in self].count(True) != 1:
            check += 1

        # The inverse must be in the set.
        if [op.inverse() in self for op in self].count(True) != len(self):
            check += 2

        # The product of two members must be in the set.
        op_prods = [op1 * op2 for op1 in self for op2 in self]

        d = self.asdict()
        for op12 in op_prods:
            if op12 not in d:
                print("op12 not in group\n %s" % str(op12))
                check += 1

        return check == 0

    def is_commutative(self) -> bool:
        """True if all operations commute with each other."""
        for op1, op2 in iuptri(self, diago=False):
            if op1 * op2 != op2 * op1: return False
        return True

    def is_abelian_group(self) -> bool:
        """True if commutative group."""
        return self.is_commutative() and self.is_group()

    def asdict(self) -> dict:
        """
        Returns a dictionary where the keys are the symmetry operations and
        the values are the indices of the operations in the iterable.
        """
        return {op: idx for idx, op in enumerate(self)}

    #def is_subset(self, other)
    #    indmap = {}
    #    for i, op in self:
    #        j = other.find(op)
    #        if j != -1: indmap[i] = j
    #    return indmap

    #def is_superset(self, other)

    @lazy_property
    def mult_table(self) -> np.ndarray:
        """
        Given a set of nsym 3x3 operations which are supposed to form a group,
        this routine constructs the multiplication table of the group.
        mtable[i,j] gives the index of the product S_i * S_j.
        """
        mtable = np.empty((len(self), len(self)), dtype=int)

        d = self.asdict()
        for i, op1 in enumerate(self):
            for j, op2 in enumerate(self):
                op12 = op1 * op2
                # Save the index of op12 in self
                try:
                    index = d[op12]
                except KeyError:
                    index = None
                mtable[i, j] = index

        return mtable

    @property
    def num_classes(self) -> int:
        """Number of classes."""
        return len(self.class_indices)

    @lazy_property
    def class_indices(self) -> list:
        """
        A class is defined as the set of distinct elements obtained by
        considering for each element, S, of the group all its conjugate
        elements X^-1 S X where X ranges over all the elements of the group.

        Returns:
            Nested list l = [cls0_indices, cls1_indices, ...] where each sublist
            contains the indices of the class. len(l) equals the number of classes.
        """
        found, class_indices = len(self) * [False], [[] for i in range(len(self))]

        num_classes = -1
        for ii, op1 in enumerate(self):
            if found[ii]: continue
            num_classes += 1

            for jj, op2 in enumerate(self):
                # Form conjugate and search it among the operations
                # that have not been found yet.
                op1_conj = op1.opconj(op2)

                for kk, op3 in enumerate(self):
                    if not found[kk] and op1_conj == op3:
                        found[kk] = True
                        class_indices[num_classes].append(kk)

        class_indices = class_indices[:num_classes + 1]
        assert sum(len(c) for c in class_indices) == len(self)
        return class_indices

    def groupby_class(self, with_inds=False):
        """
        Iterate over the operations grouped in symmetry classes.

        Args:
            with_inds: If True, [op0, op1, ...], [ind_op0, ind_op1, ...] is returned.
        """
        if with_inds:
            for indices in self.class_indices:
                yield [self[i] for i in indices], indices
        else:
            for indices in self.class_indices:
                yield [self[i] for i in indices]


class AbinitSpaceGroup(OpSequence):
    """
    Container storing the space group symmetries.
    """

    def __init__(self, spgid, symrel, tnons, symafm, has_timerev, inord="C"):
        """
        Args:
            spgid (int): space group number (from 1 to 232, 0 if cannot be specified).
            symrel: (nsym,3,3) array with the rotational part of the symmetries in real
                space (reduced coordinates are assumed, see also `inord` for the order.
            tnons: (nsym,3) array with fractional translation in reduced coordinates.
            symafm: (nsym) array with +1 for Ferromagnetic symmetry and -1 for AFM
            has_timerev: True if time-reversal symmetry is included.
            inord: storage order of mat in symrel[:]. If inord == "F", mat.T is stored
                as matrices are always stored in C-order. Use inord == "F" if you have
                read symrel from an external file produced by abinit.

        .. note::

            All the arrays are stored in C-order. Use as_fortran_arrays to extract data
            that can be passes to Fortran routines.
        """
        self.spgid = spgid
        assert 233 > self.spgid >= 0

        # Time reversal symmetry.
        self._has_timerev = has_timerev
        self._time_signs = [+1, -1] if self.has_timerev else [+1]

        self._symrel, self._tnons, self._symafm = list(map(np.asarray, (symrel, tnons, symafm)))

        if len(self.symrel) != len(self.tnons) or len(self.symrel) != len(self.symafm):
            raise ValueError("symrel, tnons and symafm must have equal shape[0]")

        inord = inord.upper()
        assert inord in ["F", "C"]
        if inord == "F":
            # Fortran to C.
            for isym in range(len(self.symrel)):
                self._symrel[isym] = self._symrel[isym].T

        self._symrec = self._symrel.copy()
        for isym in range(len(self.symrel)):
            self._symrec[isym] = mati3inv(self.symrel[isym], trans=True)

        all_syms = []
        for time_sign in self._time_signs:
            for isym in range(len(self.symrel)):
                all_syms.append(SymmOp(rot_r=self.symrel[isym],
                                       tau=self.tnons[isym],
                                       time_sign=time_sign,
                                       afm_sign=self.symafm[isym],
                                       rot_g=self.symrec[isym]))
        self._ops = tuple(all_syms)

    @classmethod
    def from_ncreader(cls, r, inord="F"):
        """
        Builds the object from a netcdf reader
        """
        kptopt = int(r.read_value("kptopt", default=1))
        symrel = r.read_value("reduced_symmetry_matrices")

        return cls(spgid=r.read_value("space_group"),
                   symrel=symrel,
                   tnons=r.read_value("reduced_symmetry_translations"),
                   symafm=r.read_value("symafm"),
                   has_timerev=has_timrev_from_kptopt(kptopt),
                   inord=inord)

    #@classmethod
    #def from_file(cls, ncfile, inord="F"):
    #    """
    #    Initialize the object from a Netcdf file.
    #    """
    #    from abipy.iotools import as_etsfreader
    #    r, closeit = as_etsfreader(ncfile)
    #    new = cls.from_ncreader(r)
    #    if closeit:
    #        file.close()

    #    return new

    @classmethod
    def from_structure(cls, structure, has_timerev=True, symprec=1e-5, angle_tolerance=5):
        """
        Takes a |Structure| object. Uses spglib to perform various symmetry finding operations.

        Args:
            structure: |Structure| object.
            has_timerev: True is time-reversal symmetry is included.
            symprec: Tolerance for symmetry finding.
            angle_tolerance: Angle tolerance for symmetry finding.

        .. warning::

            AFM symmetries are not supported.
        """
        # Call spglib to get the list of symmetry operations.
        spga = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)
        data = spga.get_symmetry_dataset()
        symrel = data["rotations"]

        return cls(spgid=data["number"],
                   symrel=symrel,
                   tnons=data["translations"],
                   symafm=len(symrel) * [1],
                   has_timerev=has_timerev,
                   inord="C")

    def __repr__(self) -> str:
        return "spgid: %d, num_spatial_symmetries: %d, has_timerev: %s, symmorphic: %s" % (
            self.spgid, self.num_spatial_symmetries, self.has_timerev, self.is_symmorphic)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = ["spgid: %d, num_spatial_symmetries: %d, has_timerev: %s, symmorphic: %s" % (
            self.spgid, self.num_spatial_symmetries, self.has_timerev, self.is_symmorphic)]
        app = lines.append

        if verbose > 1:
            for op in self.symmops(time_sign=+1):
                app(str(op))

        return "\n".join(lines)

    @lazy_property
    def is_symmorphic(self) -> bool:
        """True if there's at least one operation with non-zero fractional translation."""
        return any(op.is_symmorphic for op in self)

    @property
    def has_timerev(self) -> bool:
        """True if time-reversal symmetry is present."""
        return self._has_timerev

    @property
    def symrel(self):
        """
        [nsym, 3, 3] int array with symmetries in reduced coordinates of the direct lattice.
        """
        return self._symrel

    @property
    def tnons(self):
        """
        [nsym, 3] float array with fractional translations in reduced coordinates of the direct lattice.
        """
        return self._tnons

    @property
    def symrec(self):
        """
        [nsym, 3, 3] int array with symmetries in reduced coordinates of the reciprocal lattice.
        """
        return self._symrec

    @property
    def symafm(self):
        """[nsym] int array with +1 if FM or -1 if AFM symmetry."""
        return self._symafm

    @property
    def num_spatial_symmetries(self) -> int:
        fact = 2 if self.has_timerev else 1
        return int(len(self) / fact)

    @property
    def afm_symmops(self):
        """Tuple with antiferromagnetic symmetries."""
        return self.symmops(time_sign=None, afm_sign=-1)

    @property
    def fm_symmops(self):
        """Tuple of ferromagnetic symmetries."""
        return self.symmops(time_sign=None, afm_sign=+1)

    def symmops(self, time_sign=None, afm_sign=None):
        """
        Args:
            time_sign: If specified, only symmetries with time-reversal sign time_sign are returned.
            afm_sign: If specified, only symmetries with anti-ferromagnetic part afm_sign are returned.

        returns:
            tuple of :class:`SymmOp` instances.
        """
        symmops = []
        for sym in self._ops:
            gotit = True

            if time_sign:
                assert time_sign in self._time_signs
                gotit = gotit and sym.time_sign == time_sign

            if afm_sign:
                assert afm_sign in [-1,+1]
                gotit = gotit and sym.afm_sign == afm_sign

            if gotit:
                symmops.append(sym)

        return tuple(symmops)

    def symeq(self, k1_frac_coords, k2_frac_coords, atol=None):
        """
        Test whether two k-points in fractional coordinates are symmetry equivalent
        i.e. if there's a symmetry operations TO (including time-reversal T, if present)
        such that::

            TO(k1) = k2 + G0

        Return: namedtuple with::

            isym: The index of the symmetry operation such that TS(k1) = k2 + G0
                Set to -1 if k1 and k2 are not related by symmetry.
            op: Symmetry operation.
            g0: numpy vector.
        """
        for isym, sym in enumerate(self):
            sk_coords = sym.rotate_k(k1_frac_coords, wrap_tows=False)
            if issamek(sk_coords, k2_frac_coords, atol=atol):
                g0 = sym.rotate_k(k1_frac_coords) - k2_frac_coords
                return dict2namedtuple(isym=isym, op=self[isym], g0=g0)

        return dict2namedtuple(isym=-1, op=None, g0=None)

    def find_little_group(self, kpoint):
        """
        Find the little group of the kpoint.

        Args:
            kpoint: Accept vector with the reduced coordinates or :class:`Kpoint` object.

        Returns:
            :class:`LittleGroup` object.
        """
        if hasattr(kpoint, "frac_coords"):
            frac_coords = kpoint.frac_coords
        else:
            frac_coords = np.reshape(kpoint, (3))

        to_spgrp, g0vecs = [], []

        # Exclude AFM operations.
        for isym, symmop in enumerate(self.fm_symmops):
            is_same, g0 = symmop.preserve_k(frac_coords)
            if is_same:
                to_spgrp.append(isym)
                g0vecs.append(g0)

        # List with the symmetry operation that preserve the kpoint.
        k_symmops = [self[i] for i in to_spgrp]
        return LittleGroup(kpoint, k_symmops, g0vecs)

    def get_spglib_hall_number(self, symprec=1e-5):
        """
        Uses spglib.get_hall_number_from_symmetry to determine the hall number
        based on the symmetry operations. Useful when the space group number
        is not available, but the symmetries are (e.g. the DDB file)

        Args:
            symprec: distance tolerance in fractional coordinates (not the standard
                in cartesian coordinates). See spglib docs for more details.

        Returns:
            int: the hall number.
        """
        return spglib.get_hall_number_from_symmetry(self.symrel, self.tnons, symprec=symprec)


# FIXME To maintain backward compatibility.
SpaceGroup = AbinitSpaceGroup


class LittleGroup(OpSequence):

    def __init__(self, kpoint, symmops, g0vecs):
        """
        k_symmops, g0vecs, indices

        k_symmops is a tuple with the symmetry operations that preserve the k-point i.e. Sk = k + G0
        g0vecs is the tuple for G0 vectors for each operation in k_symmops
        """
        self.kpoint = kpoint
        self._ops = symmops
        self.g0vecs = np.reshape(g0vecs, (-1, 3))
        assert len(self.symmops) == len(self.g0vecs)

        # Find the point group of k so that we know how to access the Bilbao database.
        # (note that operations are in reciprocal space, afm and time_reversal are taken out
        krots = np.array([o.rot_g for o in symmops if not o.has_timerev])
        self.kgroup = LatticePointGroup(krots)

    @lazy_property
    def is_symmorphic(self) -> bool:
        """True if there's at least one operation with non-zero fractional translation."""
        return any(op.is_symmorphic for op in self)

    @property
    def symmops(self):
        return self._ops

    @lazy_property
    def on_bz_border(self) -> bool:
        """
        True if the k-point is on the border of the BZ.
        """
        frac_coords = np.array(self.kpoint)
        kreds = wrap_to_ws(frac_coords)
        diff = np.abs(np.abs(kreds) - 0.5)
        return np.any(diff < 1e-8)

    def iter_symmop_g0(self):
        for symmop, g0 in zip(self.symmops, self.g0vecs):
            yield symmop, g0

    def __repr__(self):
        return "Kpoint Group: %s, Kpoint: %s" % (self.kgroup, self.kpoint)

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation of little group."""
        lines = ["Kpoint-group: %s, Kpoint: %s, Symmorphic: %s" % (self.kgroup, self.kpoint, self.is_symmorphic)]
        app = lines.append
        app(" ")

        # Add character_table from Bilbao database.
        bilbao_ptgrp = bilbao_ptgroup(self.kgroup.sch_symbol)
        app(bilbao_ptgrp.to_string(verbose=verbose))
        app("")

        # Write warning if non-symmorphic little group with k-point at zone border.
        if self.is_symmorphic and self.on_bz_border:
            app("WARNING: non-symmorphic little group with k at zone-border.")
            app("Electronic states cannot be classified with this character table.")

        return "\n".join(lines)

    #def iter_symmop_g0_byclass(self):


class LatticePointGroup(OpSequence):

    def __init__(self, rotations):
        rotations = np.reshape(rotations, (-1, 3, 3))
        self._ops = [LatticeRotation(rot) for rot in rotations]

        # Call spglib to get the Herm symbol.
        # (symbol, pointgroup_number, transformation_matrix)
        herm_symbol, ptg_num, trans_mat = spglib.get_pointgroup(rotations)
        # Remove blanks from C string.
        self.herm_symbol = herm_symbol.strip()
        #print(self.herm_symbol, ptg_num, trans_mat)

        if self.sch_symbol is None:
            raise ValueError("Cannot detect point group symbol! Got sch_symbol = %s" % self.sch_symbol)

    #@classmethod
    #def from_vectors(cls, vectors)

    def __repr__(self):
        return "%s: %s, %s (%d)" % (self.__class__.__name__, self.herm_symbol, self.sch_symbol, self.spgid)

    def __str__(self):
        return "%s, %s (%d)" % (self.herm_symbol, self.sch_symbol, self.spgid)

    @property
    def sch_symbol(self) -> str:
        """Schoenflies symbol"""
        return herm2sch(self.herm_symbol)

    @property
    def spgid(self) -> int:
        """ID in the space group table."""
        return sch2spgid(self.sch_symbol)


class LatticeRotation(Operation):
    """
    This object defines a pure rotation of the lattice (proper, improper, mirror symmetry)
    that is a rotation which is compatible with a lattice. The rotation matrix is
    expressed in reduced coordinates, therefore its elements are integers.

    See:
        http://xrayweb2.chem.ou.edu/notes/symmetry.html#rotation

    .. note::

        This object is immutable and therefore we do not inherit from |numpy-array|.
    """
    _E3D = np.identity(3,  int)

    def __init__(self, mat):
        self.mat = np.asarray(mat, dtype=int)
        self.mat.shape = (3, 3)

    def _find_order_and_rootinv(self):
        """
        Returns the order of the rotation and if self is a root of the inverse.
        """
        order, root_inv = None, 0
        for ior in range(1, 7):
            rn = self ** ior

            if rn.isE:
                order = ior
                break

            if rn.isI:
                root_inv = ior

        if order is None:
            raise ValueError("LatticeRotation is not a root of unit!")

        return order, root_inv

    def __repr__(self):
        return self.name

    #def __str__(self):
    #    lines = "Rotation: " + str(self.order) + ", versor: " + str(self.versor) + ",
    #    lines.append(str(self.mat))
    #    return "\n".join(lines)

    # operator protocol.
    def __eq__(self, other):
        return np.allclose(self.mat, other.mat)

    def __mul__(self, other):
        return self.__class__(np.matmul(self.mat, other.mat))

    def __hash__(self):
        return int(8 * self.trace + 4 * self.det)

    def inverse(self):
        """
        Invert an orthogonal 3x3 matrix of INTEGER elements.
        Note use of integer arithmetic. Raise ValueError if not invertible.
        """
        return self.__class__(mati3inv(self.mat, trans=False))

    @lazy_property
    def isE(self):
        """True if it is the identity"""
        return np.allclose(self.mat, self._E3D)
    # end operator protocol.

    # Implement the unary arithmetic operations (+, -)
    def __pos__(self):
        return self

    def __neg__(self):
        return self.__class__(-self.mat)

    def __pow__(self, intexp, modulo=1):
        if intexp == 0: return self.__class__(self._E3D)
        if intexp > 0: return self.__class__(np.linalg.matrix_power(self.mat, intexp))
        if intexp == -1: return self.inverse()
        if intexp < 0: return self.__pow__(-intexp).inverse()
        raise TypeError("type %s is not supported in __pow__" % type(intexp))

    @property
    def order(self):
        """Order of the rotation."""
        try:
            return self._order
        except AttributeError:
            self._order, self._root_inv = self._find_order_and_rootinv()
            return self._order

    @property
    def root_inv(self):
        try:
            return self._root_inv
        except AttributeError:
            self._order, self._root_inv = self._find_order_and_rootinv()
            return self._root_inv

    @lazy_property
    def det(self):
        """Return the determinant of a symmetry matrix mat[3,3]. It must be +-1"""
        return _get_det(self.mat)

    @lazy_property
    def trace(self):
        """The trace of the rotation matrix"""
        return self.mat.trace()

    @lazy_property
    def is_proper(self):
        """True if proper rotation"""
        return self.det == 1

    @lazy_property
    def isI(self):
        """True if self is the inversion operation."""
        return np.allclose(self.mat, -self._E3D)

    @lazy_property
    def name(self):
        # Sign of the determinant (only if improper)
        name = "-" if self.det == -1 else ""
        name += str(self.order)
        # Root of inverse?
        name += "-" if self.root_inv != 0 else "+"

        return name

    #@property
    #def rottype(self):
    #    """
    #    Receive a 3x3 orthogonal matrix and reports its type:
    #        1 Identity
    #        2 Inversion
    #        3 Proper rotation of an angle <> 180 degrees
    #        4 Proper rotation of 180 degrees
    #        5 Mirror symmetry
    #        6 Improper rotation
    #    """
    #    # Treat identity and inversion first
    #    if self.isE: return 1
    #    if self.isI: return 2
    #
    #    if self.isproper: # Proper rotation
    #        t = 3 # try angle != 180
    #        #det180 = get_sym_det(rot + self._E3D)
    #        if (self + identity).det == 0: t = 4 # 180 rotation
    #    else:
    #        # Mirror symmetry or Improper rotation
    #        t = 6
    #        #detmirror = get_sym_det(rot - self._E3D)
    #        if (self - identity).det == 0:
    #            t = 5 # Mirror symmetry if an eigenvalue is 1

    #    return t


# TODO: Need to find an easy way to map classes in internal database
# onto classes computed by client code when calculation has been done
# with non-conventional settings (spglib?)
class Irrep:
    """
    This object represents an irreducible representation.

    .. attributes::

        traces: all_traces[nsym]. The trace of each irrep.
        character: character[num_classes]
    """
    def __init__(self, name, dim, mats, class_range):
        """
        Args:
            name:  Name of the irreducible representation.
            dim: Dimension of the irreducible representation.
            mats: Array of shape [nsym,dim,dim] with the irreducible
                representations of the group. mats are packed in classes.
            class_range: List of tuples, each tuple gives the start and stop index for the class.
                e.g. [(0, 2), (2,4), (4,n)]
        """
        self.name = name
        self._mats = np.reshape(np.array(mats), (-1, dim, dim))

        self.traces = [m.trace() for m in self.mats]

        self.class_range = class_range
        self.nclass = len(class_range)

        # Compute character table.
        character = self.nclass * [None]
        for icls, (start, stop) in enumerate(self.class_range):
            t0 = self.traces[start]
            isok = all(t0 == self.traces[i] for i in range(start, stop))
            character[icls] = t0

        self._character = character

    @property
    def mats(self):
        return self._mats

    @property
    def character(self):
        return self._character

    #@lazy_property
    #def dataframe(self):


def bilbao_ptgroup(sch_symbol: str):
    """
    Returns an instance of :class:`BilbaoPointGroup` from a string with the point group symbol
    or a number with the spacegroup ID.
    """
    sch_symbol = any2sch(sch_symbol)

    from abipy.core.irrepsdb import _PTG_IRREPS_DB
    entry = _PTG_IRREPS_DB[sch_symbol].copy()
    entry.pop("nclass")
    entry["sch_symbol"] = sch_symbol

    return BilbaoPointGroup(**entry)


class BilbaoPointGroup:
    """
    A :class:`BilbaoPointGroup` is a :class:`Pointgroup` with irreducible representations
    """
    def __init__(self, sch_symbol, rotations, class_names, class_range, irreps):
        # Rotations are grouped in classes.
        self.sch_symbol = sch_symbol
        self.rotations = np.reshape(rotations, (-1, 3, 3))
        self.class_names = class_names
        self.nclass = len(class_names)

        # List of tuples, each tuple gives the start and stop index for the class.
        # e.g. [(0, 2), (2,4), (4,n)]
        self.class_range = class_range
        self.class_len = [stop - start for start, stop in class_range]

        # The number of irreps must equal the number of classes.
        assert len(irreps) == self.nclass
        self.irreps, self.irreps_by_name = [], {}
        for name, d in irreps.items():
            mats = d["matrices"]
            assert len(mats) == self.num_rots
            irrep = Irrep(name, d["dim"], mats, class_range=self.class_range)
            self.irreps.append(irrep)
            self.irreps_by_name[name] = irrep

    @property
    def herm_symbol(self) -> str:
        """Hermann-Mauguin symbol."""
        return herm2sch(self.sch_symbol)

    @property
    def spgid(self) -> int:
        """ID in the space group table."""
        return sch2spgid(self.sch_symbol)

    @property
    def num_rots(self) -> int:
        """Number of rotations."""
        return len(self.rotations)

    @property
    def num_irreps(self) -> int:
        """Number of irreducible representations."""
        return len(self.irreps)

    @property
    def irrep_names(self) -> list:
        """List with the names of the irreps."""
        return list(self.irreps_by_name.keys())

    @lazy_property
    def character_table(self) -> pd.DataFrame:
        """
        Dataframe with irreps.
        """
        # Caveat: class names are not necessarly unique --> use np.stack
        import pandas as pd
        name_mult = [name + " [" + str(mult) + "]" for (name, mult) in zip(self.class_names, self.class_len)]
        columns = ["name"] + name_mult

        stack = np.stack([irrep.character for irrep in self.irreps])
        index = [irrep.name for irrep in self.irreps]
        df = pd.DataFrame(stack, columns=name_mult, index=index)
        df.index.name = "Irrep"
        df.columns.name = self.sch_symbol

        # TODO
        #print(df)
        # Convert complex --> real if all entries in a colums are real.
        #for k in name_mult:
        #    if np.all(np.isreal(df[k].values)):
        #        #df[k] = df[k].values.real
        #        df[k] = df[k].astype(float)

        return df

    def to_string(self, verbose=0) -> str:
        """
        Return string with the character_table
        """
        return self.character_table.to_string()

    #def decompose(self, character):
    #   od = collections.OrderedDict()
    #   for irrep in self.irreps:
    #       irrep.name
    #       irrep.character
    #   return od

    #def show_irrep(self, irrep_name):
    #    """Show the mapping rotation --> irrep mat."""
    #    irrep = self.irreps_by_name[irrep_name]

    #def irrep_from_character(self, character, rotations, tol=None):
    #    """
    #    Main entry point for client code.
    #    This routine receives a character computed from the user and finds the
    #    irreducible representation.
    #    """

    #def map_rotclasses(self, rotations_in_classes)
    #def map_rotation(self, rotations_in_classes)

    def auto_test(self):
        """
        Perform internal consistency check. Return 0 if success
        """
        #print("rotations\n", self.rotations)
        rot_group = LatticePointGroup(self.rotations)
        if not rot_group.is_group():
            print("rotations do not form a group!")
            return 1

        # Symmetries should be ordered in classes.
        # Here we recompute the classes by calling rot_group.class_indices.
        # We then sort the indices and we compare the results with the ref data stored in the Bilbao database.
        calc_class_inds = [sorted(l) for l in rot_group.class_indices]
        #print(calc_class_inds)
        assert len(calc_class_inds) == len(self.class_range)

        for calc_inds, ref_range in zip(calc_class_inds, self.class_range):
            ref_inds = list(range(ref_range[0], ref_range[1]))
            if calc_inds != ref_inds:
                print("Rotations are not ordered in classes.", calc_inds, ref_inds)
                return 2

        # Do we have a representation of the Group?
        mult_table = rot_group.mult_table
        max_err = 0.0

        for idx1, rot1 in enumerate(rot_group):
            for idx2, rot2 in enumerate(rot_group):
                idx_prod = mult_table[idx1, idx2]
                for irrep in self.irreps:
                    mat_prod = np.dot(irrep.mats[idx1], irrep.mats[idx2])
                    err = (mat_prod - irrep.mats[idx_prod]).max()
                    max_err = max(max_err, abs(err))

        if max_err > 1e-5:
            print("Irreps do not form a representation of the group, max_err: ", max_err)
            return 3

        # TODO
        # Test orthogonality theorem

        # Test the orthogonality relation of traces.
        max_err = 0.0
        for (ii, jj), (irp1, irp2) in iuptri(self.irreps, with_inds=True):
            trac1, trac2 = irp1.traces, irp2.traces
            err = np.vdot(trac1, trac2) / self.num_rots
            if ii == jj: err -= 1.0
            max_err = max(max_err, abs(err))

        if max_err > 1e-5:
            print("Error in orthogonality relation of traces: ", max_err)
            return 4

        # Success.
        return 0


# Schoenflies, Hermann-Mauguin, spgid
_PTG_IDS = [
    ("C1" , "1",     1),
    ("Ci" , "-1",    2),
    ("C2" , "2",     3),
    ("Cs" , "m",     6),
    ("C2h", "2/m",   10),
    ("D2" , "222",   16),
    ("C2v", "mm2",   25),
    ("D2h", "mmm",   47),
    ("C4" , "4",     75),
    ("S4" , "-4",    81),
    ("C4h", "4/m",   83),
    ("D4" , "422",   89),
    ("C4v", "4mm",   99),
    ("D2d", "-42m",  111),
    ("D4h", "4/mmm", 123),
    ("C3" , "3",     143),
    ("C3i", "-3",    147),
    ("D3" , "32",    149),
    ("C3v", "3m",    156),
    ("D3d", "-3m",   162),
    ("C6" , "6",     168),
    ("C3h", "-6",    174),
    ("C6h", "6/m",   175),
    ("D6" , "622",   177),
    ("C6v", "6mm",   183),
    ("D3h", "-6m2",  189),
    ("D6h", "6/mmm", 191),
    ("T"  , "23",    195),
    ("Th" , "m-3",   200),
    ("O"  , "432",   207),
    ("Td" , "-43m",  215),
    ("Oh" , "m-3m",  221),
]

_SCH2HERM = {t[0]: t[1] for t in _PTG_IDS}
_HERM2SCH = {t[1]: t[0] for t in _PTG_IDS}
_SPGID2SCH = {t[2]: t[0] for t in _PTG_IDS}
_SCH2SPGID = {t[0]: t[2] for t in _PTG_IDS}

sch_symbols = list(_SCH2HERM.keys())


def sch2herm(sch_symbol):
    """Convert from Schoenflies to Hermann-Mauguin."""
    return _SCH2HERM.get(sch_symbol, None)


def sch2spgid(sch_symbol):
    """Convert from Schoenflies to the space group id."""
    return _SCH2SPGID.get(sch_symbol, None)


def herm2sch(herm_symbol):
    """Convert from Hermann-Mauguin to Schoenflies."""
    return _HERM2SCH.get(herm_symbol, None)


def spgid2sch(spgid):
    """Return the Schoenflies symbol from the space group identifier."""
    return _SPGID2SCH.get(spgid, None)


def any2sch(obj):
    """Convert string or int to Schoenflies symbol. Returns None if invalid input"""
    if is_string(obj):
        if obj in sch_symbols:
            return obj
        else:
            # Try Hermann-Mauguin
            return herm2sch(obj)
    else:
        # Spacegroup ID?
        return spgid2sch(obj)



#!/usr/bin/env python
"""
Script to perform several types of calculations with ASE and ML potentials.
"""
from __future__ import annotations

import sys
import os
import json
import click
import numpy as np
import abipy.ml.aseml as aseml
import abipy.tools.cli_parsers as cli

from functools import wraps
from time import time
from abipy.core.structure import Structure
from abipy.tools.printing import print_dataframe

ASE_OPTIMIZERS = aseml.ase_optimizer_cls("__all__")

DEFAULT_NN = "chgnet"


def _get_atoms_from_filepath(filepath):
    if filepath.startswith("__mp-"):
        print(f"Fetching structure for mp-id {filepath[2:]} from the materials project database.")
        return aseml.get_atoms(Structure.from_mpid(filepath[2:]))

    return aseml.get_atoms(filepath)


def _get_structure_from_filepath(filepath) -> Structure:
    if filepath.startswith("__mp-"):
        print(f"Fetching structure for mp-id {filepath[2:]} from the materials project database.")
        structure = Structure.from_mpid(filepath[2:])
    else:
        structure = Structure.from_file(filepath)

    return structure


def set_default(ctx, param, filepath):
    """
    To have config file for a single command:
    Based on https://stackoverflow.com/questions/46358797/python-click-supply-arguments-and-options-from-a-configuration-file
    """
    from abipy.tools.iotools import yaml_safe_load_path
    if os.path.exists(filepath):
        print("Reading CLI options from:", filepath)
        config = yaml_safe_load_path(filepath)
        print("Config options:\n", config)
        ctx.default_map = config
    return filepath


def herald(f):
    @wraps(f)
    def wrapper(*args, **kw):
        verbose = kw.get("verbose", 0)
        #print('func:%r args:[%r, %r]' % (f.__name__, args, kw))
        #raise ValueError()
        if verbose > 3:
            print('func:%r args:[%r, %r]' % (f.__name__, args, kw))
            print(f.__doc__, end=2*"\n")
            print("Command line options:")
            print(json.dumps(kw, indent=4), end="\n")

        # IMPORTANT: Set OMP_NUM_THREADS to 1 if env variable is not defined.
        num_threads = cli.fix_omp_num_threads()

        t_start = time()
        exit_code = f(*args, **kw)
        t_end = time()
        print('\n%s command completed in %2.4f sec\n' % (f.__name__, t_end - t_start))
        return exit_code

    return wrapper


def add_constraint_opts(f):
    """Add CLI options to constrain atoms."""
    def mk_cbk(type):
        def callback(ctx, param, value):
            if value is None: return None
            return [type(s) for s in value.split()]
        return callback

    f = click.option("--fix-inds", "-fi", type=str, default=None, show_default=True,
                    callback=mk_cbk(int),
                    help='Fix atoms by indices e.g. `--fix-inds "0 1"` to fix the first two atoms.')(f)
    f = click.option("--fix-symbols", "-fs", type=str, default=None, show_default=True,
                    callback=mk_cbk(str),
                    help='Fix atoms by chemical symbols e.g. `--fix-symbols "C O"`')(f)
    return f


def add_relax_opts(f):
    """Add CLI options for structural relaxations with ASE."""
    # fmax (float): total force tolerance for relaxation convergence.
    # Here fmax is a sum of force and stress forces. Defaults to 0.1.
    f = click.option("--relax-mode", "-r", default="ions", show_default=True, type=click.Choice(["no", "ions", "cell"]),
                     help='Relaxation mode.')(f)
    f = click.option("--fmax", default=0.01, type=float, show_default=True,
                     help='Stopping criterion in eV/A')(f)
    f = click.option("--pressure", default=0.0, type=float, show_default=True, help='Scalar pressure')(f)
    f = click.option("--steps", default=500, type=int, show_default=True,
                     help="Max number of steps for ASE relaxation.")(f)
    f = click.option("--optimizer", "-o", default="BFGS", show_default=True, type=click.Choice(ASE_OPTIMIZERS),
                     help="ASE optimizer class.")(f)
    return f


def add_neb_opts(f):
    """Add CLI options for NEB calculations with ASE."""
    f = click.option("--nimages", "-n", default=14, type=click.IntRange(3, None), show_default=True,
                     help='Number of NEB images including initial/final points. Must be >= 3')(f)
    f = click.option("--relax-mode", "-r", default="ions", show_default=True, type=click.Choice(["no", "ions", "cell"]),
            help="Relax initial and final structure. Use `cell` to relax ions and cell, " +
                 "`ions` to relax atomic positions only, `no` to disable relaxation")(f)
    f = click.option("--fmax", default=0.03, type=float, show_default=True, help='Stopping criterion.')(f)
    f = click.option("--pressure", default=0.0, type=float, show_default=True, help='Scalar pressure')(f)
    f = click.option("--optimizer", "-o", default="BFGS", show_default=True, type=click.Choice(ASE_OPTIMIZERS),
                     help="ASE optimizer class.")(f)
    f = click.option("--neb-method", "-m", default="aseneb", type=click.Choice(aseml.ASENEB_METHODS),
                     show_default=True, help="ASE NEB method")(f)
    f = click.option("--climb", "-c", is_flag=True, help="Use a climbing image (default is no climbing image).")(f)
    return f


def add_nprocs_opt(f):
    """Add CLI options for multiprocessing."""
    f = click.option("-np", "--nprocs", default=-1, type=int, show_default=True,
                    help='Number of processes in multiprocessing pool. -1 to let Abipy select it automatically.')(f)
    return f


def add_workdir_verbose_opts(f):
    """Add workdir and verbose options to CLI subcommand."""
    f = click.option("--workdir", "-w", default=None, type=str,
                      help="Working directory. If not specified, a temporary directory is created.")(f)
    f = click.option('-v', '--verbose', count=True, help="Verbosity level")(f)
    return f


def add_nn_name_opt(f):
    """Add CLI options to select the NN potential."""
    f = click.option("--nn-name", "-nn", default=DEFAULT_NN, show_default=True,
                     help=f"ML potential to be used. Supported values are: {aseml.CalcBuilder.ALL_NN_TYPES}")(f)
    #f = click.option("--dftd3", , default="no", show_default=True,
    #                 help=f"Activate DFD3.")(f)
    return f


def add_nn_names_opt(f):
    """Add CLI options to select multiple NN potentials."""
    f = click.option("-nns", '--nn-names', type=str, multiple=True, show_default=True,
                    help='ML potentials to use.', default=[DEFAULT_NN])(f)
    return f


def _get_nn_names(nn_names: list[str]) -> list[str]:
    """
    Pre-processing of nn-names option.

    --nn-names all                   --> return all NN names installed in this env.
    --nn-names all-alignn-m3gnet     --> return all NN names except alignn and m3gnet.
    --nn-names chgnet-               --> return all NN names except chgnet.
    """
    if "all" in nn_names:
        # Return all possibilities.
        nn_installed, nn_versions = aseml.get_installed_nn_names(verbose=0, printout=False)
        return nn_installed

    if any(n.startswith("all-") for n in nn_names):
        # --nn-names all-alignn-m3gnet     --> return all NN names except alignn and m3gnet
        assert len(nn_names) == 1
        skip_names = nn_names[0].replace("all-", "").split("-")
        return [s for s in aseml.CalcBuilder.ALL_NN_TYPES if s not in skip_names]

    if any(n.endswith("-") for n in nn_names):
        # --nn-names chgnet-               --> return all NN names except chgnet.
        skip_names = [n[:-2] for n in nn_names if n.endswith("-")]
        return [s for s in aseml.CalcBuilder.ALL_NN_TYPES if s not in skip_names]

    return nn_names


@click.group()
@click.pass_context
@click.option("--seaborn", "-sns", default=None, show_default=True,
              help='Use seaborn settings. Accept value defining context in ("paper", "notebook", "talk", "poster").')
def main(ctx, seaborn):
    """Script to perform calculations with ML potentials."""
    ctx.ensure_object(dict)

    if seaborn:
        # Activate seaborn settings for plots
        import seaborn as sns
        sns.set(context=seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_name_opt
@add_relax_opts
@add_constraint_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_relax.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def relax(ctx, filepath, nn_name,
          relax_mode, fmax, pressure, steps, optimizer,
          fix_inds, fix_symbols,
          workdir, verbose):
    """
    Structural relaxation with ASE and ML potential.

    Usage example:

    \b
        abiml.py relax FILE --fmax 0.01 -r cell --optimizer FIRE -w OUT_DIR
        abiml.py relax FILE --fix-inds "0 3" --fix-symbols "Si O"

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.
    or a string such as __mp-134 to fetch the structure from the MP database.

    To change the ML potential, use e.g.:

        abiml.py relax -nn m3gnet [...]
    """
    atoms = _get_atoms_from_filepath(filepath)
    aseml.fix_atoms(atoms, fix_inds=fix_inds, fix_symbols=fix_symbols)

    ml_relaxer = aseml.MlRelaxer(atoms, relax_mode, fmax, pressure, steps, optimizer,
                                 nn_name, verbose, workdir, prefix="_abiml_relax_")

    print(ml_relaxer.to_string(verbose=verbose))
    ml_relaxer.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_workdir_verbose_opts
@click.option('--config', default='abiml_abinit_relax.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def abinit_relax(ctx, filepath,
                 workdir, verbose):
    """
    Interact with ABINIT in hybrid relaxation mode.
    """
    ml_relaxer = aseml.MlRelaxer.from_abinit_yaml_file(filepath)
    print(ml_relaxer.to_string(verbose=verbose))
    ml_relaxer.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_name_opt
@click.option('--temperature', "-t", default=600, type=float, show_default=True, help='Temperature in Kelvin')
@click.option('--timestep', "-ts", default=1, type=float, show_default=True, help='Timestep in fs.')
@click.option('--steps', "-s", default=1000, type=int, show_default=True, help='Number of timesteps.')
@click.option('--loginterval', "-l", default=100, type=int, show_default=True, help='Interval for record the log.')
@click.option('--ensemble', "-e", default="nvt", show_default=True,
              type=click.Choice(["nvt", "npt", "npt_berendsen"]), help='Ensemble e.g. nvt, npt.')
@add_constraint_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_md.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def md(ctx, filepath, nn_name,
       temperature, timestep, steps, loginterval, ensemble,
       fix_inds, fix_symbols,
       workdir, verbose):
    """
    MD simulation with ASE and ML potential.

    Usage example:

    \b
        abiml.py md FILE --temperature 1200 --timestep 2 --steps 5000 --workdir OUT_DIR
        abiml.py md FILE --fix-inds "0 3" --fix-symbols "Si O"

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.

    To change the ML potential, use e.g.:

        abiml.py md -nn m3gnet [...]

    To restart a MD run, use --workdir to specify a pre-existent directory.
    """
    # See https://github.com/materialsvirtuallab/m3gnet#molecular-dynamics
    atoms = aseml.get_atoms(filepath)
    aseml.fix_atoms(atoms, fix_inds=fix_inds, fix_symbols=fix_symbols)

    ml_md = aseml.MlMd(atoms, temperature, timestep, steps, loginterval, ensemble, nn_name, verbose,
                       workdir, prefix="_abiml_md_")
    print(ml_md.to_string(verbose=verbose))
    ml_md.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepaths", nargs=2, type=str)
@add_nn_name_opt
@add_neb_opts
@add_constraint_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_neb.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def neb(ctx, filepaths, nn_name,
        nimages, relax_mode, fmax, pressure, optimizer, neb_method, climb,
        fix_inds, fix_symbols,
        workdir, verbose
    ):
    """
    NEB calculation with ASE and ML potential.

    Usage example:

    \b
        abiml.py neb START_FILE END_FILE --nimages 6 --fmax=0.05 --optimizer FIRE -w OUT_DIR
        abiml.py neb START_FILE END_FILE --neb-method improvedtangent --climb
        abiml.py neb START_FILE END_FILE --fix-inds "0 3" --fix-symbols "Si O"

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.

    To change the ML potential, use e.g.:

        abiml.py neb -nn m3gnet [...]
    """
    initial_atoms = aseml.get_atoms(filepaths[0])
    aseml.fix_atoms(initial_atoms, fix_inds=fix_inds, fix_symbols=fix_symbols)
    final_atoms = aseml.get_atoms(filepaths[1])
    aseml.fix_atoms(final_atoms, fix_inds=fix_inds, fix_symbols=fix_symbols)

    ml_neb = aseml.MlNeb(initial_atoms, final_atoms,
                         nimages, neb_method, climb, optimizer, relax_mode, fmax, pressure,
                         nn_name, verbose, workdir, prefix="_abiml_neb_")
    print(ml_neb.to_string(verbose=verbose))
    ml_neb.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepaths", nargs=-1, type=str) # , help="Files with structures")
@add_nn_name_opt
@add_neb_opts
@add_constraint_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_mneb.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def mneb(ctx, filepaths, nn_name,
         nimages, relax_mode, fmax, pressure, optimizer, neb_method, climb,
         fix_inds, fix_symbols,
         workdir, verbose):
    """
    Multi-NEB calculation with ASE and ML potential.

    Usage example:

    \b
        abiml.py mneb FILE1 FILE2 FILE2 ... --nimages 6 --fmax=0.05 -w OUT_DIR
        abiml.py mneb FILE1 FILE2 FILE2 ... --neb-method improvedtangent --climb
        abiml.py mneb FILE1 FILE2 FILE2 ... --fix-inds "0 3" --fix-symbols "Si O"

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.

    To change the ML potential, use e.g.:

        abiml.py mneb -nn m3gnet [...]
    """
    # Fix atoms
    atoms_list = [aseml.get_atoms(p) for p in filepaths]
    for atoms in atoms_list:
        aseml.fix_atoms(atoms, fix_inds=fix_inds, fix_symbols=fix_symbols)

    mneb = aseml.MultiMlNeb(atoms_list, nimages, neb_method, climb, optimizer, relax_mode, fmax, pressure,
                            nn_name, verbose, workdir, prefix="_abiml_mneb_")
    print(mneb.to_string(verbose=verbose))
    mneb.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_names_opt
@click.option("--supercell", "-s", nargs=3, type=int, default=(2, 2, 2), show_default=True, help="Supercell dimensions.")
@click.option("--distance", "-d", type=float, show_default=True, default=0.01, help="Displacement distance in Ang.")
#@click.option('--asr', type=int, default=2, show_default=True, help="Restore the acoustic sum rule on the interatomic force constants.")
#@click.option('--dipdip', type=int, default=1, show_default=True, help="Treatment of dipole-dipole interaction.")
@click.option('--line-density', "-ld", default=20, type=float, show_default=True, help="Line density to generate the q-path for PH bands.")
@click.option('--qppa', "-qppa", default=None, type=float, show_default=True, help="q-points per atom to generate the q-mesh for PH DOS.")
@add_relax_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_ph.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def ph(ctx, filepath, nn_names,
       supercell, distance,
       #asr, dipdip,
       line_density, qppa,
       relax_mode, fmax, pressure, steps, optimizer,
       workdir, verbose):
    """
    Use phonopy and ML potential to compute phonons.

    Usage example:

    \b
        abiml.py ph FILE --distance 0.03 --supercell 2 2 2

    where `FILE` provides the crystalline structure
    or a string such as __mp-134 to fetch the structure from the MP database.

    To specify the list of ML potential, use e.g.:

        abiml.py ddb -nn-names m3gnet --nn-names chgnet [...]

    To use all NN potentials supported, use:

        -nn-names all [...]
    """
    structure = _get_structure_from_filepath(filepath)

    from abipy.ml.ml_phonopy import MlPhonopy
    supercell = np.eye(3) * np.array(supercell)

    nn_names = _get_nn_names(nn_names)
    ml_ph = MlPhonopy(structure, supercell,
                      distance,
                      # asr, dipdip,
                      line_density, qppa,
                      relax_mode, fmax, pressure, steps, optimizer, nn_names,
                      verbose, workdir, prefix="_abiml_ph_",
                      )
    print(ml_ph.to_string(verbose=verbose))
    ml_ph.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("ddb_filepath", type=str)
@add_nn_names_opt
@click.option("--supercell", "-s", nargs=3, type=int, default=(-1, -1, -1), show_default=True, help="Supercell. If < 0, supercell is taken from DDB ngqpt.")
@click.option("--distance", "-d", type=float, show_default=True, default=0.01, help="Displacement distance in Ang.")
@click.option('--asr', type=int, default=2, show_default=True, help="Restore the acoustic sum rule on the interatomic force constants.")
@click.option('--dipdip', type=int, default=1, show_default=True, help="Treatment of dipole-dipole interaction.")
@click.option('--line-density', "-ld", default=20, type=float, show_default=True, help="Line density to generate the q-path for PH bands.")
@click.option('--qppa', "-qppa", default=None, type=float, show_default=True, help="q-points per atom to generate the q-mesh for PH DOS.")
@add_relax_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_phddb.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def phddb(ctx, ddb_filepath, nn_names,
          supercell, distance, asr, dipdip, line_density, qppa,
          relax_mode, fmax, pressure, steps, optimizer,
          workdir, verbose):
    """
    Use phonopy and ML potential to compute phonons and compare with DDB.

    Usage example:

    \b
        abiml.py phddb DDB_FILE --distance 0.03 --dipdip 0 --supercell 2 2 2

    where `DDB_FILE` is an Abinit DDB file
    or a string such as __mp-134 to fetch the DDB from the MP database.

    To specify the list of ML potential, use e.g.:

        abiml.py phddb -nn-names m3gnet --nn-names chgnet [...]

    To use all NN potentials supported, use:

        -nn-names all [...]
    """
    if ddb_filepath.startswith("__mp-"):
        print(f"Fetching DDB for mp-id {ddb_filepath[2:]} from the materials project database.")
        from abipy.dfpt.ddb import DdbFile
        with DdbFile.from_mpid(ddb_filepath[2:]) as ddb:
            ddb_filepath = ddb.filepath

    from abipy.ml.ml_phonopy import MlPhonopyWithDDB
    if any(s <= 0 for s in supercell):
        supercell = None
    else:
        supercell = np.eye(3) * np.array(supercell)

    nn_names = _get_nn_names(nn_names)
    ml_phddb = MlPhonopyWithDDB(ddb_filepath,
                                distance, asr, dipdip, line_density, qppa,
                                relax_mode, fmax, pressure, steps, optimizer, nn_names,
                                verbose, workdir, prefix="_abiml_phddb_",
                                supercell=supercell,
                                )
    print(ml_phddb.to_string(verbose=verbose))
    ml_phddb.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_name_opt
@click.option("--max-ns", "-m", default=100, type=int, show_default=True, help='Max number of structures')
@add_relax_opts
@add_workdir_verbose_opts
@click.option('--config', default='abiml_order.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def order(ctx, filepath, nn_name,
          max_ns, relax_mode, fmax, pressure, steps, optimizer, workdir, verbose):
    """
    Generate ordered structures from CIF with partial occupancies.

    Usage example:

    \b
        abiml.py order FILE --max-ns 10 --relax cell

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.

    Based on: https://matgenb.materialsvirtuallab.org/2013/01/01/Ordering-Disordered-Structures.html
    """
    ml_orderer = aseml.MlOrderer(filepath, max_ns, optimizer, relax_mode, fmax, pressure,
                                 steps, nn_name, verbose, workdir, prefix="_abiml_order_")
    print(ml_orderer.to_string(verbose=verbose))
    ml_orderer.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_name_opt
@click.option("-isite", "--isite", required=True,
               help='Index of atom to displace or string with the chemical element to be added to input structure.')
@click.option("--mesh", type=int, default=4, show_default=True, help='Mesh size along the smallest cell size.')
@add_relax_opts
@add_nprocs_opt
@add_workdir_verbose_opts
@click.option('--config', default='abiml_scan_relax.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def scan_relax(ctx, filepath, nn_name,
               isite, mesh,
               relax_mode, fmax, pressure, steps, optimizer,
               nprocs,
               workdir, verbose
               ):
    """
    Generate 3D mesh of (nx,ny,nz) initial positions and perform multiple relaxations
    in which all atoms are fixed except the one initially placed at the mesh point.

    Usage example:

    \b
        abiml.py scan-relax FILE -isite 0 --mesh 4  # Move first atom in the structure
        abiml.py scan-relax FILE -isite H           # Add H to the structure read from FILE.

    where `FILE` is any file supported by abipy/pymatgen e.g. netcdf files, Abinit input, POSCAR, xsf, etc.

    To change the ML potential, use e.g.:

        abiml.py scan-relax -nn m3gnet [...]
    """
    structure = Structure.from_file(filepath)

    from abipy.ml.relax_scanner import RelaxScanner
    scanner = RelaxScanner(structure, isite, mesh, nn_name,
                           relax_mode=relax_mode, fmax=fmax, steps=steps, verbose=verbose,
                           optimizer_name=optimizer, pressure=pressure,
                           workdir=workdir, prefix="_abiml_scan_relax_")
    print(scanner)
    scanner.run(nprocs=nprocs)
    return 0


@main.command()
@herald
@click.pass_context
@click.argument('filepaths', type=str, nargs=-1)
@add_nn_names_opt
@click.option("--traj_range", type=str, show_default=True,
              help="Trajectory range e.g. `5` to select the first 5 iterations, `1:4` to select steps 1,2,3. `1:4:2 for 1,3",
              default=None)
@click.option("-e", '--exposer', default="mpl", show_default=True, type=click.Choice(["mpl", "panel"]),
              help='Plotting backend: mpl for matplotlib, panel for web-based, None to disable plotting')
@add_nprocs_opt
@add_workdir_verbose_opts
@click.option('--config', default='abiml_validate.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def validate(ctx, filepaths,
             nn_names,
             traj_range,
             exposer,
             nprocs,
             workdir, verbose
            ):
    """
    Compare ab-initio energies, forces, and stresses with ml-computed ones.

    usage example:

    \b
        abiml.py validate FILE --nn-names matgl --nn-names chgnet

    where `FILE` can be a HIST.nc, a vasprun.xml or an ASE extended XYZ file.
    """
    traj_range = cli.range_from_str(traj_range)
    nn_names = _get_nn_names(nn_names)
    ml_comp = aseml.MlValidateWithAbinitio(filepaths, nn_names, traj_range, verbose, workdir, prefix="_abiml_validate_")
    print(ml_comp)
    c = ml_comp.run(nprocs=nprocs)

    if exposer != "None":
        from abipy.tools.plotting import Exposer
        with_stress = True
        with Exposer.as_exposer(exposer, title=" ".join(os.path.basename(p) for p in filepaths)) as e:
            e(c.plot_energies(show=False))
            e(c.plot_forces(delta_mode=True, show=False))
            e(c.plot_energies_traj(delta_mode=True, show=False))
            e(c.plot_energies_traj(delta_mode=False, show=False))
            if with_stress:
                e(c.plot_stresses(delta_mode=True, show=False))
            e(c.plot_forces_traj(delta_mode=True, show=False))
            e(c.plot_stress_traj(delta_mode=True, show=False))

    return 0


@main.command()
@herald
@click.pass_context
@click.option('-v', '--verbose', count=True, help="Verbosity level")
def show(ctx, verbose):
    """
    Show the NN potentials installed in the environment.
    """
    installed, versions = aseml.get_installed_nn_names(verbose=verbose, printout=True)
    return 0 if installed else 1


@main.command()
@herald
@click.pass_context
@click.option("-nns", '--nn-names', type=str, multiple=True, show_default=True,
              help='ML potentials to install.', default=["all"])
@click.option('-U', '--update', is_flag=True, default=False, show_default=True, help="Update packages.")
@click.option('-v', '--verbose', count=True, help="Verbosity level")
def install(ctx, nn_names, update, verbose):
    """
    Install NN potentials in the environment using pip.
    """
    aseml.install_nn_names(nn_names=nn_names, update=update, verbose=verbose)
    installed, versions = aseml.get_installed_nn_names(verbose=verbose, printout=True)

    return 0 if installed else 1


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@click.option("-nns", '--nn-names', type=str, multiple=True, show_default=True,
              help='ML potentials to compare.', default=["all"])
@click.option('--num-tests', "-n", default=20, type=int, show_default=True, help='Number of configurations to generate.')
@click.option("--rattle", default=0.2, type=float, show_default=True, help="Displace atoms randomly with this stdev.")
@click.option("-srv", "--stdev-rvol", default=0.1, type=float, show_default=True,
              help="Scale volumes randomly around input v0 with stdev: v0 * value")
@add_workdir_verbose_opts
@click.option('--config', default='abiml_compare.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def compare(ctx, filepath, nn_names,
            num_tests, rattle, stdev_rvol,
            workdir, verbose
         ):
    """
    Compare different neural networks.
    """
    atoms = _get_atoms_from_filepath(filepath)
    nn_names = _get_nn_names(nn_names)
    ml_comp = aseml.MlCompareNNs(atoms, nn_names, num_tests, rattle, stdev_rvol, verbose, workdir, prefix="_abiml_compare_")
    print(ml_comp.to_string(verbose=verbose))
    ase_comp = ml_comp.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("filepath", type=str)
@add_nn_name_opt
@add_workdir_verbose_opts
@click.option('--config', default='abiml_gs.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def gs(ctx, filepath, nn_name,
       workdir, verbose,
       ):
    """
    Compute ground-state properties and magnetic moments with ML potential(s).
    """
    atoms = _get_atoms_from_filepath(filepath)
    gs = aseml.GsMl(atoms, nn_name, verbose, workdir, prefix="_abiml_gs_")
    gs.run()
    return 0


@main.command()
@herald
@click.pass_context
@click.argument("elements", nargs=-1, type=str)
@add_nn_names_opt
@add_workdir_verbose_opts
@click.option('--config', default='abiml_cwf_eos.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
def cwf_eos(ctx, elements, nn_names,
            workdir, verbose
            ):
    """
    Compute CWF EOS with ML potentials.
    """
    nn_names = _get_nn_names(nn_names)
    if "all" in elements:
        if len(elements) != 1:
            raise ValueError(f"When all is used for elements len(elements) should be 1 while {elements=}")
        from ase.data import chemical_symbols
        elements = [chemical_symbols[Z] for Z in range(1, 96+1)]

    ml_cwf_eos = aseml.MlCwfEos(elements, nn_names, verbose, workdir, prefix="_abiml_cwf_eos_")
    print(ml_cwf_eos.to_string(verbose=verbose))
    ml_cwf_eos.run()
    return 0


#@main.command()
#@herald
#@click.pass_context
#@click.argument('filepaths', type=str, nargs=-1)
#@add_nn_name_opt
#@add_workdir_verbose_opts
#@click.option('--config', default='abiml_train.yml', type=click.Path(), callback=set_default, is_eager=True, expose_value=False)
#def train(ctx, filepaths,
#          nn_name,
#          #nprocs,
#          workdir, verbose
#          ):
#    """
#    Train a ML potential using the trajectory stored on FILE.
#
#    usage example:
#
#    \b
#        abiml.py train FILE --nn-names matgl
#
#    where `FILE` can be either a _HIST.nc or a vasprun.xml FILE.
#    """
#    if nn_name == "matgl":
#        from abipy.ml.matgl import MatglSystem
#        s = MatglSystem(filepaths, workdir, verbose)
#
#    elif nn_name == "chgnet":
#        from abipy.ml.chgnet import ChgnetSystem
#        s = ChgnetSystem(filepaths, workdir, verbose)
#
#    else:
#        raise ValueError(f"Unsupported {nn_name=}")
#
#    s.train()
#    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Phonon fatbands
===============

This example shows how to plot the phonon fatbands of AlAs.
See tutorial/lesson_rf2.html
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the PHBST file produced by anaddb and get the phonon bands.
with abiopen(abidata.ref_file("trf2_5.out_PHBST.nc")) as ncfile:
    phbands = ncfile.phbands

# Plot the phonon band structure with matplotlib.
phbands.plot_fatbands(title="AlAs phonon fatbands without LO-TO splitting")

# For the plotly version, use:
phbands.plotly_fatbands(title="AlAs phonon fatbands without LO-TO splitting")

# Provide the PHDOS file path produced by anaddb.
# sphinx_gallery_thumbnail_number = 2
phdos_path = abidata.ref_file("trf2_5.out_PHDOS.nc")

# Plot the phonon band structure + PJDOS with matplotlib.
phbands.plot_fatbands(units="Thz", phdos_file=phdos_path, title="AlAs phonon fatbands with PJDOS")

# For the plotly version, use:
phbands.plotly_fatbands(units="Thz", phdos_file=phdos_path, title="AlAs phonon fatbands with PJDOS")

# Plot contributions to the phonon displacement at the Gamma point grouped by atom type.
phbands.plot_phdispl(qpoint=(0, 0, 0), units="meV", title=r"phonon displacements at $\Gamma$")

# Plot (orthonormal) eigenvectors instead of displacements.
phbands.plot_phdispl(qpoint=(0, 0, 0), use_eigvec=True, units="meV", title=r"phonon eigenvectors at $\Gamma$")

# Decompose contributions along the three Cartesian directions.
phbands.plot_phdispl_cartdirs(qpoint=(0, 0, 0), units="cm-1")
